@@ -35,6 +35,9 @@
 
 #include "cache_varnishd.h"
 #include <stdio.h>
+#include <stdlib.h>
+
+#include "common/heritage.h"
 
 #include "vct.h"
 #include "vend.h"
@@ -58,6 +61,8 @@ const char H__Status[]	= "\010:status:";
 const char H__Proto[]	= "\007:proto:";
 const char H__Reason[]	= "\010:reason:";
 
+static char * via_hdr;
+
 /*--------------------------------------------------------------------
  * Perfect hash to rapidly recognize headers from tbl/http_headers.h
  * which have non-zero flags.
@@ -140,23 +145,23 @@ http_hdr_flags(const char *b, const char *e)
 	unsigned u;
 	struct http_hdrflg *retval;
 
-	if (e == NULL)
-		return(NULL);
-	assert(e > b);
+	if (b == NULL || e == NULL)
+		return (NULL);
+	assert(b <= e);
 	u = (unsigned)(e - b);
 	assert(b + u == e);
 	if (u < 2 || u > 19)		// MIN_WORD_LENGTH & MAX_WORD_LENGTH
-		return(NULL);
+		return (NULL);
 	if (u > 3)
 		u += http_asso_values[((const uint8_t*)b)[3]];
 	if (u > 38)			// MAX_HASH_VALUE
-		return(NULL);
+		return (NULL);
 	retval = &http_hdrflg[u];
 	if (retval->hdr == NULL)
-		return(NULL);
+		return (NULL);
 	if (!http_hdr_at(retval->hdr + 1, b, e - b))
-		return(NULL);
-	return(retval);
+		return (NULL);
+	return (retval);
 }
 
 /*--------------------------------------------------------------------*/
@@ -178,9 +183,18 @@ http_init_hdr(char *hdr, int flg)
 void
 HTTP_Init(void)
 {
+	struct vsb *vsb;
 
 #define HTTPH(a, b, c) http_init_hdr(b, c);
 #include "tbl/http_headers.h"
+
+	vsb = VSB_new_auto();
+	AN(vsb);
+	VSB_printf(vsb, "1.1 %s (Varnish/" PACKAGE_BRANCH ")",
+	    heritage.identity);
+	AZ(VSB_finish(vsb));
+	REPLACE(via_hdr, VSB_data(vsb));
+	VSB_destroy(&vsb);
 }
 
 /*--------------------------------------------------------------------
@@ -415,7 +429,7 @@ http_PutField(struct http *to, int field, const char *string)
 	p = WS_Copy(to->ws, string, -1);
 	if (p == NULL) {
 		http_fail(to);
-		VSLb(to->vsl, SLT_LostHeader, "%s", string);
+		VSLbs(to->vsl, SLT_LostHeader, TOSTRAND(string));
 		return;
 	}
 	http_SetH(to, field, p);
@@ -535,7 +549,8 @@ http_CollectHdrSep(struct http *hp, hdr_t hdr, const char *sep)
 			x = Tlen(hp->hd[f]);
 			if (b + x >= e) {
 				http_fail(hp);
-				VSLb(hp->vsl, SLT_LostHeader, "%s", hdr + 1);
+				VSLbs(hp->vsl, SLT_LostHeader,
+				    TOSTRAND(hdr + 1));
 				WS_Release(hp->ws, 0);
 				return;
 			}
@@ -557,7 +572,7 @@ http_CollectHdrSep(struct http *hp, hdr_t hdr, const char *sep)
 
 		if (b + lsep + x >= e) {
 			http_fail(hp);
-			VSLb(hp->vsl, SLT_LostHeader, "%s", hdr + 1);
+			VSLbs(hp->vsl, SLT_LostHeader, TOSTRAND(hdr + 1));
 			WS_Release(hp->ws, 0);
 			return;
 		}
@@ -1132,6 +1147,8 @@ http_isfiltered(const struct http *fm, unsigned u, unsigned how)
 
 	if (fm->hdf[u] & HDF_FILTER)
 		return (1);
+	if (u < HTTP_HDR_FIRST)
+		return (0);
 	e = strchr(fm->hd[u].b, ':');
 	if (e == NULL)
 		return (0);
@@ -1441,7 +1458,7 @@ http_CopyHome(const struct http *hp)
 		p = WS_Copy(hp->ws, hp->hd[u].b, l + 1L);
 		if (p == NULL) {
 			http_fail(hp);
-			VSLb(hp->vsl, SLT_LostHeader, "%s", hp->hd[u].b);
+			VSLbs(hp->vsl, SLT_LostHeader, TOSTRAND(hp->hd[u].b));
 			return;
 		}
 		hp->hd[u].b = p;
@@ -1457,7 +1474,7 @@ http_SetHeader(struct http *to, const char *header)
 
 	CHECK_OBJ_NOTNULL(to, HTTP_MAGIC);
 	if (to->nhd >= to->shd) {
-		VSLb(to->vsl, SLT_LostHeader, "%s", header);
+		VSLbs(to->vsl, SLT_LostHeader, TOSTRAND(header));
 		http_fail(to);
 		return;
 	}
@@ -1477,6 +1494,20 @@ http_ForceHeader(struct http *to, hdr_t hdr, const char *val)
 	http_PrintfHeader(to, "%s %s", hdr + 1, val);
 }
 
+void
+http_AppendHeader(struct http *to, hdr_t hdr, const char *val)
+{
+	const char *old;
+
+	http_CollectHdr(to, hdr);
+	if (http_GetHdr(to, hdr, &old)) {
+		http_Unset(to, hdr);
+		http_PrintfHeader(to, "%s %s, %s", hdr + 1, old, val);
+	} else {
+		http_PrintfHeader(to, "%s %s", hdr + 1, val);
+	}
+}
+
 void
 http_PrintfHeader(struct http *to, const char *fmt, ...)
 {
@@ -1511,14 +1542,14 @@ http_TimeHeader(struct http *to, const char *fmt, vtim_real now)
 
 	CHECK_OBJ_NOTNULL(to, HTTP_MAGIC);
 	if (to->nhd >= to->shd) {
-		VSLb(to->vsl, SLT_LostHeader, "%s", fmt);
+		VSLbs(to->vsl, SLT_LostHeader, TOSTRAND(fmt));
 		http_fail(to);
 		return;
 	}
 	p = WS_Alloc(to->ws, strlen(fmt) + VTIM_FORMAT_SIZE);
 	if (p == NULL) {
 		http_fail(to);
-		VSLb(to->vsl, SLT_LostHeader, "%s", fmt);
+		VSLbs(to->vsl, SLT_LostHeader, TOSTRAND(fmt));
 		return;
 	}
 	strcpy(p, fmt);
@@ -1526,6 +1557,13 @@ http_TimeHeader(struct http *to, const char *fmt, vtim_real now)
 	http_SetH(to, to->nhd++, p);
 }
 
+const char *
+http_ViaHeader(void)
+{
+
+	return (via_hdr);
+}
+
 /*--------------------------------------------------------------------*/
 
 void
