@@ -44,6 +44,7 @@
 #include "cache_filter.h"
 #include "cache_objhead.h"
 #include "cache_transport.h"
+#include "vcc_interface.h"
 
 #include "hash/hash_slinger.h"
 #include "http1/cache_http1.h"
@@ -72,7 +73,7 @@
 #define REQ_STEP(l, U, priv) \
     static req_state_f cnt_##l; \
     priv const struct req_step R_STP_##U[1] = {{ \
-	.name = "Fetch Step" #l, \
+	.name = "Req Step " #l, \
 	.func = cnt_##l, \
     }};
 REQ_STEPS
@@ -110,6 +111,7 @@ cnt_transport(struct worker *wrk, struct req *req)
 
 	req->doclose = http_DoConnection(req->http, SC_REQ_CLOSE);
 	if (req->doclose == SC_RX_BAD) {
+		wrk->stats->client_req_400++;
 		(void)req->transport->minimal_response(req, 400);
 		return (REQ_FSM_DONE);
 	}
@@ -166,7 +168,7 @@ Resp_Setup_Deliver(struct req *req)
 	http_PrintfHeader(h, "Age: %.0f",
 	    floor(fmax(0., req->t_prev - oc->t_origin)));
 
-	http_SetHeader(h, "Via: 1.1 varnish (Varnish/7.0)");
+	http_AppendHeader(h, H_Via, http_ViaHeader());
 
 	if (cache_param->http_gzip_support &&
 	    ObjCheckFlag(req->wrk, oc, OF_GZIPED) &&
@@ -230,17 +232,19 @@ cnt_deliver(struct worker *wrk, struct req *req)
 	    !(req->objcore->flags & OC_F_PRIVATE))
 		http_ForceHeader(req->resp, H_Accept_Ranges, "bytes");
 
+	req->t_resp = W_TIM_real(wrk);
 	VCL_deliver_method(req->vcl, wrk, req, NULL, NULL);
+
 	VSLb_ts_req(req, "Process", W_TIM_real(wrk));
 
 	assert(req->restarts <= cache_param->max_restarts);
 
-	if (wrk->handling != VCL_RET_DELIVER) {
+	if (wrk->vpi->handling != VCL_RET_DELIVER) {
 		HSH_Cancel(wrk, req->objcore, NULL);
 		(void)HSH_DerefObjCore(wrk, &req->objcore, HSH_RUSH_POLICY);
 		http_Teardown(req->resp);
 
-		switch (wrk->handling) {
+		switch (wrk->vpi->handling) {
 		case VCL_RET_RESTART:
 			req->req_step = R_STP_RESTART;
 			break;
@@ -257,7 +261,7 @@ cnt_deliver(struct worker *wrk, struct req *req)
 		return (REQ_FSM_MORE);
 	}
 
-	assert(wrk->handling == VCL_RET_DELIVER);
+	assert(wrk->vpi->handling == VCL_RET_DELIVER);
 
 	if (IS_TOPREQ(req) && RFC2616_Do_Cond(req))
 		http_PutResponse(req->resp, "HTTP/1.1", 304, NULL);
@@ -304,6 +308,7 @@ cnt_synth(struct worker *wrk, struct req *req)
 	struct vsb *synth_body;
 	ssize_t sz, szl;
 	uint8_t *ptr;
+	const char *body;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
@@ -322,13 +327,14 @@ cnt_synth(struct worker *wrk, struct req *req)
 	synth_body = VSB_new_auto();
 	AN(synth_body);
 
+	req->t_resp = W_TIM_real(wrk);
 	VCL_synth_method(req->vcl, wrk, req, NULL, synth_body);
 
 	AZ(VSB_finish(synth_body));
 
 	VSLb_ts_req(req, "Process", W_TIM_real(wrk));
 
-	if (wrk->handling == VCL_RET_FAIL) {
+	if (wrk->vpi->handling == VCL_RET_FAIL) {
 		VSB_destroy(&synth_body);
 		(void)VRB_Ignore(req);
 		(void)req->transport->minimal_response(req, 500);
@@ -338,11 +344,11 @@ cnt_synth(struct worker *wrk, struct req *req)
 		return (REQ_FSM_DONE);
 	}
 
-	if (wrk->handling == VCL_RET_RESTART &&
+	if (wrk->vpi->handling == VCL_RET_RESTART &&
 	    req->restarts > cache_param->max_restarts)
-		wrk->handling = VCL_RET_DELIVER;
+		wrk->vpi->handling = VCL_RET_DELIVER;
 
-	if (wrk->handling == VCL_RET_RESTART) {
+	if (wrk->vpi->handling == VCL_RET_RESTART) {
 		/*
 		 * XXX: Should we reset req->doclose = SC_VCL_FAILURE
 		 * XXX: If so, to what ?
@@ -352,13 +358,14 @@ cnt_synth(struct worker *wrk, struct req *req)
 		req->req_step = R_STP_RESTART;
 		return (REQ_FSM_MORE);
 	}
-	assert(wrk->handling == VCL_RET_DELIVER);
+	assert(wrk->vpi->handling == VCL_RET_DELIVER);
 
 	http_Unset(req->resp, H_Content_Length);
 	http_PrintfHeader(req->resp, "Content-Length: %zd",
 	    VSB_len(synth_body));
 
-	if (!req->doclose && http_HdrIs(req->resp, H_Connection, "close"))
+	if (req->doclose == SC_NULL &&
+	    http_HdrIs(req->resp, H_Connection, "close"))
 		req->doclose = SC_RESP_CLOSE;
 
 	/* Discard any lingering request body before delivery */
@@ -368,20 +375,26 @@ cnt_synth(struct worker *wrk, struct req *req)
 	CHECK_OBJ_NOTNULL(req->objcore, OBJCORE_MAGIC);
 	szl = -1;
 	if (STV_NewObject(wrk, req->objcore, stv_transient, 1024)) {
+		body = VSB_data(synth_body);
 		szl = VSB_len(synth_body);
 		assert(szl >= 0);
-		sz = szl;
-		if (sz > 0 &&
-		    ObjGetSpace(wrk, req->objcore, &sz, &ptr) && sz >= szl) {
-			memcpy(ptr, VSB_data(synth_body), szl);
-			ObjExtend(wrk, req->objcore, szl, 1);
-		} else if (sz > 0) {
-			szl = -1;
+		while (szl > 0) {
+			sz = szl;
+			if (! ObjGetSpace(wrk, req->objcore, &sz, &ptr)) {
+				szl = -1;
+				break;
+			}
+			if (sz > szl)
+				sz = szl;
+			szl -= sz;
+			memcpy(ptr, body, sz);
+			ObjExtend(wrk, req->objcore, sz, szl == 0 ? 1 : 0);
+			body += sz;
 		}
 	}
 
 	if (szl >= 0)
-		AZ(ObjSetU64(wrk, req->objcore, OA_LEN, szl));
+		AZ(ObjSetU64(wrk, req->objcore, OA_LEN, VSB_len(synth_body)));
 	HSH_DerefBoc(wrk, req->objcore);
 	VSB_destroy(&synth_body);
 
@@ -478,10 +491,11 @@ cnt_transmit(struct worker *wrk, struct req *req)
 
 	HSH_Cancel(wrk, req->objcore, boc);
 
-	if (!req->doclose && (req->objcore->flags & OC_F_FAILED))
+	if (req->doclose == SC_NULL && (req->objcore->flags & OC_F_FAILED)) {
 		/* The object we delivered failed due to a streaming error.
 		 * Fail the request. */
 		req->doclose = SC_TX_ERROR;
+	}
 
 	if (boc != NULL)
 		HSH_DerefBoc(wrk, req->objcore);
@@ -548,10 +562,6 @@ cnt_lookup(struct worker *wrk, struct req *req)
 		had_objhead = 1;
 	wrk->strangelove = 0;
 	lr = HSH_Lookup(req, &oc, &busy);
-	assert(wrk->strangelove >= 0);
-	if ((unsigned)wrk->strangelove >= cache_param->vary_notice)
-		VSLb(req->vsl, SLT_Notice, "vsl: High number of variants (%d)",
-		    wrk->strangelove);
 	if (lr == HSH_BUSY) {
 		/*
 		 * We lost the session to a busy object, disembark the
@@ -561,6 +571,10 @@ cnt_lookup(struct worker *wrk, struct req *req)
 		 */
 		return (REQ_FSM_DISEMBARK);
 	}
+	assert(wrk->strangelove >= 0);
+	if ((unsigned)wrk->strangelove >= cache_param->vary_notice)
+		VSLb(req->vsl, SLT_Notice, "vsl: High number of variants (%d)",
+		    wrk->strangelove);
 	if (had_objhead)
 		VSLb_ts_req(req, "Waitinglist", W_TIM_real(wrk));
 
@@ -604,7 +618,7 @@ cnt_lookup(struct worker *wrk, struct req *req)
 
 	VCL_hit_method(req->vcl, wrk, req, NULL, NULL);
 
-	switch (wrk->handling) {
+	switch (wrk->vpi->handling) {
 	case VCL_RET_DELIVER:
 		if (busy != NULL) {
 			AZ(oc->flags & OC_F_HFM);
@@ -666,7 +680,7 @@ cnt_miss(struct worker *wrk, struct req *req)
 	CHECK_OBJ_ORNULL(req->stale_oc, OBJCORE_MAGIC);
 
 	VCL_miss_method(req->vcl, wrk, req, NULL, NULL);
-	switch (wrk->handling) {
+	switch (wrk->vpi->handling) {
 	case VCL_RET_FETCH:
 		wrk->stats->cache_miss++;
 		VBF_Fetch(wrk, req, req->objcore, req->stale_oc, VBF_NORMAL);
@@ -711,7 +725,7 @@ cnt_pass(struct worker *wrk, struct req *req)
 	AZ(req->stale_oc);
 
 	VCL_pass_method(req->vcl, wrk, req, NULL, NULL);
-	switch (wrk->handling) {
+	switch (wrk->vpi->handling) {
 	case VCL_RET_FAIL:
 		req->req_step = R_STP_VCLFAIL;
 		break;
@@ -755,6 +769,7 @@ cnt_pipe(struct worker *wrk, struct req *req)
 	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
 	VSLb(bo->vsl, SLT_Begin, "bereq %u pipe", VXID(req->vsl->wid));
 	VSLb(req->vsl, SLT_Link, "bereq %u pipe", VXID(bo->vsl->wid));
+	VSLb_ts_busyobj(bo, "Start", W_TIM_real(wrk));
 	THR_SetBusyobj(bo);
 	bo->sp = req->sp;
 	SES_Ref(bo->sp);
@@ -771,16 +786,18 @@ cnt_pipe(struct worker *wrk, struct req *req)
 
 	bo->wrk = wrk;
 	if (WS_Overflowed(req->ws))
-		wrk->handling = VCL_RET_FAIL;
+		wrk->vpi->handling = VCL_RET_FAIL;
 	else
 		VCL_pipe_method(req->vcl, wrk, req, bo, NULL);
 
-	switch (wrk->handling) {
+	switch (wrk->vpi->handling) {
 	case VCL_RET_SYNTH:
 		req->req_step = R_STP_SYNTH;
 		nxt = REQ_FSM_MORE;
 		break;
 	case VCL_RET_PIPE:
+		VSLb_ts_req(req, "Process", W_TIM_real(wrk));
+		VSLb_ts_busyobj(bo, "Process", wrk->lastused);
 		if (V1P_Enter() == 0) {
 			AZ(bo->req);
 			bo->req = req;
@@ -850,25 +867,19 @@ cnt_restart(struct worker *wrk, struct req *req)
 static void v_matchproto_(req_state_f)
 cnt_recv_prep(struct req *req, const char *ci)
 {
-	const char *xff;
 
 	if (req->restarts == 0) {
 		/*
 		 * This really should be done earlier, but we want to capture
 		 * it in the VSL log.
 		 */
-		http_CollectHdr(req->http, H_X_Forwarded_For);
-		if (http_GetHdr(req->http, H_X_Forwarded_For, &xff)) {
-			http_Unset(req->http, H_X_Forwarded_For);
-			http_PrintfHeader(req->http, "X-Forwarded-For: %s, %s",
-			    xff, ci);
-		} else {
-			http_PrintfHeader(req->http, "X-Forwarded-For: %s", ci);
-		}
+		http_AppendHeader(req->http, H_X_Forwarded_For, ci);
+		http_AppendHeader(req->http, H_Via, http_ViaHeader());
 		http_CollectHdr(req->http, H_Cache_Control);
 
 		/* By default we use the first backend */
-		req->director_hint = VCL_DefaultDirector(req->vcl);
+		VRT_Assign_Backend(&req->director_hint,
+		    VCL_DefaultDirector(req->vcl));
 
 		req->d_ttl = -1;
 		req->d_grace = -1;
@@ -920,6 +931,20 @@ cnt_recv(struct worker *wrk, struct req *req)
 
 	http_VSL_log(req->http);
 
+	if (http_CountHdr(req->http0, H_Host) > 1) {
+		VSLb(req->vsl, SLT_BogoHeader, "Multiple Host: headers");
+		wrk->stats->client_req_400++;
+		(void)req->transport->minimal_response(req, 400);
+		return (REQ_FSM_DONE);
+	}
+
+	if (http_CountHdr(req->http0, H_Content_Length) > 1) {
+		VSLb(req->vsl, SLT_BogoHeader, "Multiple Content-Length: headers");
+		wrk->stats->client_req_400++;
+		(void)req->transport->minimal_response(req, 400);
+		return (REQ_FSM_DONE);
+	}
+
 	cnt_recv_prep(req, ci);
 
 	if (req->req_body_status == BS_ERROR) {
@@ -929,12 +954,12 @@ cnt_recv(struct worker *wrk, struct req *req)
 
 	VCL_recv_method(req->vcl, wrk, req, NULL, NULL);
 
-	if (wrk->handling == VCL_RET_FAIL) {
+	if (wrk->vpi->handling == VCL_RET_FAIL) {
 		req->req_step = R_STP_VCLFAIL;
 		return (REQ_FSM_MORE);
 	}
 
-	if (wrk->handling == VCL_RET_VCL && req->restarts == 0) {
+	if (wrk->vpi->handling == VCL_RET_VCL && req->restarts == 0) {
 		// Req_Rollback has happened in VPI_vcl_select
 		assert(WS_Snapshot(req->ws) == req->ws_req);
 		cnt_recv_prep(req, ci);
@@ -955,7 +980,7 @@ cnt_recv(struct worker *wrk, struct req *req)
 		return (REQ_FSM_DONE);
 	}
 
-	recv_handling = wrk->handling;
+	recv_handling = wrk->vpi->handling;
 
 	/* We wash the A-E header here for the sake of VRY */
 	if (cache_param->http_gzip_support &&
@@ -970,10 +995,10 @@ cnt_recv(struct worker *wrk, struct req *req)
 
 	VSHA256_Init(&sha256ctx);
 	VCL_hash_method(req->vcl, wrk, req, NULL, &sha256ctx);
-	if (wrk->handling == VCL_RET_FAIL)
-		recv_handling = wrk->handling;
+	if (wrk->vpi->handling == VCL_RET_FAIL)
+		recv_handling = wrk->vpi->handling;
 	else
-		assert(wrk->handling == VCL_RET_LOOKUP);
+		assert(wrk->vpi->handling == VCL_RET_LOOKUP);
 	VSHA256_Final(req->digest, &sha256ctx);
 
 	switch (recv_handling) {
@@ -1058,7 +1083,7 @@ cnt_purge(struct worker *wrk, struct req *req)
 	AZ(HSH_DerefObjCore(wrk, &boc, 1));
 
 	VCL_purge_method(req->vcl, wrk, req, NULL, NULL);
-	switch (wrk->handling) {
+	switch (wrk->vpi->handling) {
 	case VCL_RET_RESTART:
 		req->req_step = R_STP_RESTART;
 		break;
@@ -1106,7 +1131,7 @@ CNT_Embark(struct worker *wrk, struct req *req)
 		VCL_Refresh(&wrk->wpriv->vcl);
 		req->vcl = wrk->wpriv->vcl;
 		wrk->wpriv->vcl = NULL;
-		VSLb(req->vsl, SLT_VCL_use, "%s", VCL_Name(req->vcl));
+		VSLbs(req->vsl, SLT_VCL_use, TOSTRAND(VCL_Name(req->vcl)));
 	}
 
 	AN(req->vcl);
@@ -1147,6 +1172,7 @@ CNT_Request(struct req *req)
 		CHECK_OBJ_NOTNULL(wrk->wpriv, WORKER_PRIV_MAGIC);
 		CHECK_OBJ_ORNULL(wrk->wpriv->nobjhead, OBJHEAD_MAGIC);
 		CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
+		CHECK_OBJ_NOTNULL(req->doclose, STREAM_CLOSE_MAGIC);
 
 		AN(req->req_step);
 		AN(req->req_step->name);
@@ -1168,6 +1194,7 @@ CNT_Request(struct req *req)
 		VCL_TaskLeave(ctx, req->privs);
 		AN(req->vsl->wid);
 		VRB_Free(req);
+		VRT_Assign_Backend(&req->director_hint, NULL);
 		req->wrk = NULL;
 	}
 	assert(nxt == REQ_FSM_DISEMBARK || !WS_IsReserved(req->ws));
