@@ -68,7 +68,7 @@ struct lock		vcl_mtx;
 struct vcl		*vcl_active; /* protected by vcl_mtx */
 
 static struct vrt_ctx ctx_cli;
-static unsigned handling_cli;
+static struct wrk_vpi wrk_vpi_cli;
 static struct ws ws_cli;
 static uintptr_t ws_snapshot_cli;
 static struct vsl_log vsl_cli;
@@ -91,8 +91,8 @@ VCL_Bo2Ctx(struct vrt_ctx *ctx, struct busyobj *bo)
 	ctx->sp = bo->sp;
 	ctx->now = bo->t_prev;
 	ctx->ws = bo->ws;
-	ctx->handling = &bo->wrk->handling;
-	*ctx->handling = 0;
+	ctx->vpi = bo->wrk->vpi;
+	ctx->vpi->handling = 0;
 }
 
 void
@@ -101,6 +101,7 @@ VCL_Req2Ctx(struct vrt_ctx *ctx, struct req *req)
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
+	CHECK_OBJ_NOTNULL(req->doclose, STREAM_CLOSE_MAGIC);
 
 	ctx->vcl = req->vcl;
 	ctx->syntax = ctx->vcl->conf->syntax;
@@ -113,8 +114,8 @@ VCL_Req2Ctx(struct vrt_ctx *ctx, struct req *req)
 	ctx->sp = req->sp;
 	ctx->now = req->t_prev;
 	ctx->ws = req->ws;
-	ctx->handling = &req->wrk->handling;
-	*ctx->handling = 0;
+	ctx->vpi = req->wrk->vpi;
+	ctx->vpi->handling = 0;
 }
 
 /*--------------------------------------------------------------------*/
@@ -124,10 +125,9 @@ VCL_Get_CliCtx(int msg)
 {
 
 	ASSERT_CLI();
-	AZ(ctx_cli.handling);
 	INIT_OBJ(&ctx_cli, VRT_CTX_MAGIC);
-	handling_cli = 0;
-	ctx_cli.handling = &handling_cli;
+	INIT_OBJ(&wrk_vpi_cli, WRK_VPI_MAGIC);
+	ctx_cli.vpi = &wrk_vpi_cli;
 	ctx_cli.now = VTIM_real();
 	if (msg) {
 		ctx_cli.msg = VSB_new_auto();
@@ -155,7 +155,7 @@ VCL_Rel_CliCtx(struct vrt_ctx **ctx)
 
 	ASSERT_CLI();
 	assert(*ctx == &ctx_cli);
-	AN((*ctx)->handling);
+	AN((*ctx)->vpi);
 	if (ctx_cli.msg) {
 		TAKE_OBJ_NOTNULL(r, &ctx_cli.msg, VSB_MAGIC);
 		AZ(VSB_finish(r));
@@ -186,10 +186,10 @@ vcl_event_handling(VRT_CTX)
 {
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 
-	if (*ctx->handling == 0)
+	if (ctx->vpi->handling == 0)
 		return (0);
 
-	assert(*ctx->handling == VCL_RET_FAIL);
+	assert(ctx->vpi->handling == VCL_RET_FAIL);
 
 	if (ctx->method == VCL_MET_INIT)
 		return (1);
@@ -200,7 +200,7 @@ vcl_event_handling(VRT_CTX)
 	 */
 	assert(ctx->method == VCL_MET_FINI);
 
-	*ctx->handling = 0;
+	ctx->vpi->handling = 0;
 	VRT_fail(ctx, "VRT_fail() from vcl_fini{} has no effect");
 	return (0);
 }
@@ -239,8 +239,8 @@ vcl_send_event(struct vcl *vcl, enum vcl_event_e ev, struct vsb **msg)
 
 	ctx = VCL_Get_CliCtx(havemsg);
 
-	AN(ctx->handling);
-	AZ(*ctx->handling);
+	AN(ctx->vpi);
+	AZ(ctx->vpi->handling);
 	AN(ctx->ws);
 
 	ctx->vcl = vcl;
@@ -282,7 +282,7 @@ vcl_find(const char *name)
 static void
 vcl_panic_conf(struct vsb *vsb, const struct VCL_conf *conf)
 {
-	int i;
+	unsigned u;
 	const struct vpi_ii *ii;
 
 	if (PAN_dump_struct(vsb, conf, VCL_CONF_MAGIC, "conf"))
@@ -290,8 +290,8 @@ vcl_panic_conf(struct vsb *vsb, const struct VCL_conf *conf)
 	VSB_printf(vsb, "syntax = \"%u\",\n", conf->syntax);
 	VSB_cat(vsb, "srcname = {\n");
 	VSB_indent(vsb, 2);
-	for (i = 0; i < conf->nsrc; ++i)
-		VSB_printf(vsb, "\"%s\",\n", conf->srcname[i]);
+	for (u = 0; u < conf->nsrc; ++u)
+		VSB_printf(vsb, "[%u] = \"%s\",\n", u, conf->srcname[u]);
 	VSB_indent(vsb, -2);
 	VSB_cat(vsb, "},\n");
 	VSB_cat(vsb, "instances = {\n");
@@ -447,24 +447,19 @@ vcl_BackendEvent(const struct vcl *vcl, enum vcl_event_e e)
 }
 
 static void
-vcl_KillBackends(struct vcl *vcl)
+vcl_KillBackends(const struct vcl *vcl)
 {
-	struct vcldir *vdir;
+	struct vcldir *vdir, *vdir2;
 
 	CHECK_OBJ_NOTNULL(vcl, VCL_MAGIC);
-	AZ(vcl->busy);
-	assert(VTAILQ_EMPTY(&vcl->ref_list));
-	Lck_Lock(&vcl_mtx);
-	while (1) {
-		vdir = VTAILQ_FIRST(&vcl->director_list);
-		if (vdir == NULL)
-			break;
-		VTAILQ_REMOVE(&vcl->director_list, vdir, list);
-		AN(vdir->methods->destroy);
-		vdir->methods->destroy(vdir->dir);
-		vcldir_free(vdir);
-	}
-	Lck_Unlock(&vcl_mtx);
+	assert(vcl->temp == VCL_TEMP_COLD || vcl->temp == VCL_TEMP_INIT);
+	/*
+	 * Unlocked because no further directors can be added, and the
+	 * remaining ones need to be able to remove themselves.
+	 */
+	VTAILQ_FOREACH_SAFE(vdir, &vcl->director_list, list, vdir2)
+		VDI_Event(vdir->dir, VCL_EVENT_DISCARD);
+	assert(VTAILQ_EMPTY(&vcl->director_list));
 }
 
 /*--------------------------------------------------------------------*/
@@ -521,8 +516,7 @@ VCL_Close(struct vcl **vclp)
 	struct vcl *vcl;
 
 	TAKE_OBJ_NOTNULL(vcl, vclp, VCL_MAGIC);
-	assert(VTAILQ_EMPTY(&vcl->vfps));
-	assert(VTAILQ_EMPTY(&vcl->vdps));
+	assert(VTAILQ_EMPTY(&vcl->filters));
 	AZ(dlclose(vcl->dlh));
 	FREE_OBJ(vcl);
 }
@@ -690,8 +684,7 @@ vcl_load(struct cli *cli,
 	XXXAN(vcl->loaded_name);
 	VTAILQ_INIT(&vcl->director_list);
 	VTAILQ_INIT(&vcl->ref_list);
-	VTAILQ_INIT(&vcl->vfps);
-	VTAILQ_INIT(&vcl->vdps);
+	VTAILQ_INIT(&vcl->filters);
 
 	vcl->temp = VCL_TEMP_INIT;
 
@@ -953,28 +946,34 @@ vcl_cli_show(struct cli *cli, const char * const *av, void *priv)
 {
 	struct vcl *vcl;
 	int verbose = 0;
-	int i;
+	int i = 2;
+	unsigned u;
 
 	ASSERT_CLI();
 	ASSERT_VCL_ACTIVE();
 	AZ(priv);
-	if (!strcmp(av[2], "-v") && av[3] == NULL) {
-		VCLI_Out(cli, "Too few parameters");
-		VCLI_SetResult(cli, CLIS_TOOFEW);
-		return;
-	} else if (strcmp(av[2], "-v") && av[3] != NULL) {
-		VCLI_Out(cli, "Unknown options '%s'", av[2]);
+
+	if (av[i] != NULL && !strcmp(av[i], "-v")) {
+		verbose = 1;
+		i++;
+	}
+
+	if (av[i] == NULL) {
+		vcl = vcl_active;
+		AN(vcl);
+	} else {
+		vcl = vcl_find(av[i]);
+		i++;
+	}
+
+	if (av[i] != NULL) {
+		VCLI_Out(cli, "Too many parameters: '%s'", av[i]);
 		VCLI_SetResult(cli, CLIS_PARAM);
 		return;
-	} else if (av[3] != NULL) {
-		verbose = 1;
-		vcl = vcl_find(av[3]);
-	} else
-		vcl = vcl_find(av[2]);
+	}
 
 	if (vcl == NULL) {
-		VCLI_Out(cli, "No VCL named '%s'",
-		    av[3] == NULL ? av[2] : av[3]);
+		VCLI_Out(cli, "No VCL named '%s'", av[i - 1]);
 		VCLI_SetResult(cli, CLIS_PARAM);
 		return;
 	}
@@ -986,11 +985,11 @@ vcl_cli_show(struct cli *cli, const char * const *av, void *priv)
 	}
 	CHECK_OBJ_NOTNULL(vcl->conf, VCL_CONF_MAGIC);
 	if (verbose) {
-		for (i = 0; i < vcl->conf->nsrc; i++)
-			VCLI_Out(cli, "// VCL.SHOW %d %zd %s\n%s\n",
-			    i, strlen(vcl->conf->srcbody[i]),
-			    vcl->conf->srcname[i],
-			    vcl->conf->srcbody[i]);
+		for (u = 0; u < vcl->conf->nsrc; u++)
+			VCLI_Out(cli, "// VCL.SHOW %u %zd %s\n%s\n",
+			    u, strlen(vcl->conf->srcbody[u]),
+			    vcl->conf->srcname[u],
+			    vcl->conf->srcbody[u]);
 	} else {
 		VCLI_Out(cli, "%s", vcl->conf->srcbody[0]);
 	}
