@@ -35,6 +35,7 @@
 #include <stdio.h>
 
 #include "cache_varnishd.h"
+#include "cache_transport.h"
 #include "common/heritage.h"
 
 #include "vcl.h"
@@ -180,7 +181,7 @@ VRT_r_obj_reason(VRT_CTX)
 }
 
 /*--------------------------------------------------------------------
- * bool-fields (.do_*)
+ * beresp bool-fields
  */
 
 static inline int
@@ -216,7 +217,6 @@ VRT_l_beresp_##field(VRT_CTX, VCL_BOOL a)				\
 		return (0);			\
 	} while(0)
 
-#define VBERESPR0(field, str, fltchk)
 #define VBERESPR1(field, str, fltchk)					\
 VCL_BOOL								\
 VRT_r_beresp_##field(VRT_CTX)						\
@@ -227,21 +227,10 @@ VRT_r_beresp_##field(VRT_CTX)						\
 	return (ctx->bo->field);					\
 }
 
-#define VBEREQR0(field, str, fltchk)
-#define VBEREQR1(field, str, fltchk)					\
-VCL_BOOL								\
-VRT_r_bereq_##field(VRT_CTX)						\
-{									\
-	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);				\
-	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);			\
-	return (ctx->bo->field);					\
-}
-
-#define BO_FLAG(l, r, rr, rw, f, d)		\
-	VBEREQR##r(l, #l, f)			\
-	VBERESPR##rr(l, #l, f)			\
-	VBERESPW##rw(l, #l, f)
-#include "tbl/bo_flags.h"
+#define BERESP_FLAG(l, r, w, f, d)		\
+	VBERESPR##r(l, #l, f)			\
+	VBERESPW##w(l, #l, f)
+#include "tbl/beresp_flags.h"
 
 #undef VBERESPWF0
 #undef VBERESPWF1
@@ -250,8 +239,28 @@ VRT_r_bereq_##field(VRT_CTX)						\
 
 #undef VBERESPRF0
 #undef VBERESPRF1
-#undef VBERESPR0
 #undef VBERESPR1
+
+/*--------------------------------------------------------------------
+ * bereq bool-fields
+ */
+
+#define VBEREQR0(field, str)
+#define VBEREQR1(field, str)						\
+VCL_BOOL								\
+VRT_r_bereq_##field(VRT_CTX)						\
+{									\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);				\
+	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);			\
+	return (ctx->bo->field);					\
+}
+
+#define BEREQ_FLAG(l, r, w, d)		\
+	VBEREQR##r(l, #l)
+#include "tbl/bereq_flags.h"
+
+#undef VBEREQR0
+#undef VBEREQR1
 /*--------------------------------------------------------------------*/
 
 VCL_BOOL
@@ -504,13 +513,22 @@ VRT_r_req_##nm(VRT_CTX)							\
 	return (ctx->req->elem);					\
 }
 
-REQ_VAR_L(backend_hint, director_hint, VCL_BACKEND,)
 REQ_VAR_R(backend_hint, director_hint, VCL_BACKEND)
+
 REQ_VAR_L(ttl, d_ttl, VCL_DURATION, if (!(arg>0.0)) arg = 0;)
 REQ_VAR_R(ttl, d_ttl, VCL_DURATION)
 REQ_VAR_L(grace, d_grace, VCL_DURATION, if (!(arg>0.0)) arg = 0;)
 REQ_VAR_R(grace, d_grace, VCL_DURATION)
 
+VCL_VOID
+VRT_l_req_backend_hint(VRT_CTX, VCL_BACKEND be)
+{
+
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
+	VRT_Assign_Backend(&ctx->req->director_hint, be);
+}
+
 /*--------------------------------------------------------------------*/
 
 VCL_VOID
@@ -519,7 +537,7 @@ VRT_l_bereq_backend(VRT_CTX, VCL_BACKEND be)
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
-	ctx->bo->director_req = be;
+	VRT_Assign_Backend(&ctx->bo->director_req, be);
 }
 
 VCL_BACKEND
@@ -625,6 +643,18 @@ VRT_r_bereq_retries(VRT_CTX)
 	return (ctx->bo->retries);
 }
 
+/*--------------------------------------------------------------------*/
+
+VCL_STRING
+VRT_r_req_transport(VRT_CTX)
+{
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
+
+	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
+	CHECK_OBJ_NOTNULL(ctx->req->transport, TRANSPORT_MAGIC);
+	return (ctx->req->transport->name);
+}
+
 /*--------------------------------------------------------------------
  * In exp.*:
  *	t_origin is absolute
@@ -698,6 +728,25 @@ VRT_DO_EXP_R(beresp, ctx->bo->fetch_objcore, keep, 0)
 
 /*lint -restore */
 
+// XXX more assertions?
+#define VRT_DO_TIME_R(which, where, field)				\
+									\
+VCL_TIME								\
+VRT_r_##which##_time(VRT_CTX)						\
+{									\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);				\
+	AN((ctx)->where);						\
+									\
+	return ((ctx)->where->field);					\
+}
+
+VRT_DO_TIME_R(req, req, t_req)
+VRT_DO_TIME_R(req_top, req->top->topreq, t_req)
+VRT_DO_TIME_R(resp, req, t_resp)
+VRT_DO_TIME_R(bereq, bo, t_first)
+VRT_DO_TIME_R(beresp, bo, t_resp)
+VRT_DO_TIME_R(obj, req->objcore, t_origin)
+
 /*--------------------------------------------------------------------
  */
 
@@ -937,17 +986,29 @@ VRT_r_resp_do_esi(VRT_CTX)
 #define VRT_BODY_L(which)					\
 VCL_VOID							\
 VRT_l_##which##_body(VRT_CTX, enum lbody_e type,		\
-    const char *str, VCL_STRANDS s)				\
+    const char *str, VCL_BODY body)				\
 {								\
 	int n;							\
 	struct vsb *vsb;					\
+	VCL_STRANDS s;						\
+	VCL_BLOB b;						\
 								\
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);			\
+	AN(body);						\
 	CAST_OBJ_NOTNULL(vsb, ctx->specific, VSB_MAGIC);	\
-	assert(type == LBODY_SET || type == LBODY_ADD);		\
-	if (type == LBODY_SET)					\
+	if (type == LBODY_SET_STRING || type == LBODY_SET_BLOB)	\
 		VSB_clear(vsb);					\
+	if (type == LBODY_SET_BLOB || type == LBODY_ADD_BLOB) {	\
+		AZ(str);					\
+		b = body;					\
+		VSB_bcat(vsb, b->blob, b->len);			\
+		return;						\
+	}							\
 	if (str != NULL)					\
 		VSB_cat(vsb, str);				\
+	assert(type == LBODY_SET_STRING ||			\
+	    type == LBODY_ADD_STRING);				\
+	s = body;						\
 	for (n = 0; s != NULL && n < s->n; n++)			\
 		if (s->p[n] != NULL)				\
 			VSB_cat(vsb, s->p[n]);			\
@@ -1011,8 +1072,7 @@ VRT_l_sess_##x(VRT_CTX, VCL_DURATION d)		\
 {							\
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);		\
 	CHECK_OBJ_NOTNULL(ctx->sp, SESS_MAGIC);		\
-	if (d < 0.0)					\
-		d = 0.0;				\
+	d = vmax(d, 0.0);				\
 	setter;						\
 	ctx->sp->x = d;					\
 }							\
