@@ -42,7 +42,7 @@
  *
  * In our case, any worker thread which we wake up comes from the idle queue,
  * where it put itself under the mutex, releasing that mutex implicitly via
- * Lck_CondWait() (which calls some variant of pthread_cond_wait). So we avoid
+ * Lck_CondWaitUntil() (which calls some variant of pthread_cond_wait). So we avoid
  * additional mutex contention knowing that any worker thread on the idle queue
  * is blocking on the cv.
  *
@@ -89,7 +89,6 @@ wrk_bgthread(void *arg)
 	struct worker wrk;
 	struct worker_priv wpriv[1];
 	struct VSC_main_wrk ds;
-	void *r;
 
 	CAST_OBJ_NOTNULL(bt, arg, BGTHREAD_MAGIC);
 	THR_SetName(bt->name);
@@ -97,15 +96,11 @@ wrk_bgthread(void *arg)
 	INIT_OBJ(&wrk, WORKER_MAGIC);
 	INIT_OBJ(wpriv, WORKER_PRIV_MAGIC);
 	wrk.wpriv = wpriv;
+	// bgthreads do not have a vpi member
 	memset(&ds, 0, sizeof ds);
 	wrk.stats = &ds;
 
-	r = bt->func(&wrk, bt->priv);
-
-	if (! cache_shutdown)
-		WRONG("BgThread terminated");
-
-	return (r);
+	return (bt->func(&wrk, bt->priv));
 }
 
 void
@@ -132,6 +127,7 @@ WRK_Thread(struct pool *qp, size_t stacksize, unsigned thread_workspace)
 	struct VSC_main_wrk ds;
 	unsigned char ws[thread_workspace];
 	struct worker_priv wpriv[1];
+	unsigned char vpi[vpi_wrk_len];
 
 	AN(qp);
 	AN(stacksize);
@@ -149,6 +145,8 @@ WRK_Thread(struct pool *qp, size_t stacksize, unsigned thread_workspace)
 	AZ(pthread_cond_init(&w->cond, NULL));
 
 	WS_Init(w->aws, "wrk", ws, thread_workspace);
+	VPI_wrk_init(w, vpi, sizeof vpi);
+	AN(w->vpi);
 
 	VSL(SLT_WorkThread, 0, "%p start", w);
 
@@ -218,8 +216,8 @@ pool_reserve(void)
 		if (cache_param->wthread_reserve < lim)
 			lim = cache_param->wthread_reserve;
 	}
-	if (lim < TASK_QUEUE__END)
-		return (TASK_QUEUE__END);
+	if (lim < TASK_QUEUE_RESERVE)
+		return (TASK_QUEUE_RESERVE);
 	return (lim);
 }
 
@@ -233,7 +231,7 @@ pool_getidleworker(struct pool *pp, enum task_prio prio)
 
 	CHECK_OBJ_NOTNULL(pp, POOL_MAGIC);
 	Lck_AssertHeld(&pp->mtx);
-	if (pp->nidle > (pool_reserve() * prio / TASK_QUEUE__END)) {
+	if (pp->nidle > (pool_reserve() * prio / TASK_QUEUE_RESERVE)) {
 		pt = VTAILQ_FIRST(&pp->idle_queue);
 		if (pt == NULL)
 			AZ(pp->nidle);
@@ -333,11 +331,12 @@ Pool_Task(struct pool *pp, struct pool_task *task, enum task_prio prio)
 		return (0);
 	}
 
-	/*
-	 * queue limits only apply to client threads - all other
-	 * work is vital and needs do be done at the earliest
+	/* Vital work is always queued. Only priority classes that can
+	 * fit under the reserve capacity are eligible to queuing.
 	 */
-	if (!TASK_QUEUE_CLIENT(prio) ||
+	if (prio >= TASK_QUEUE_RESERVE) {
+		retval = -1;
+	} else if (!TASK_QUEUE_LIMITED(prio) ||
 	    pp->lqueue + pp->nthr < cache_param->wthread_max +
 	    cache_param->wthread_queue_limit) {
 		pp->nqueued++;
@@ -391,8 +390,8 @@ Pool_Work_Thread(struct pool *pp, struct worker *wrk)
 		Lck_Lock(&pp->mtx);
 		reserve = pool_reserve();
 
-		for (i = 0; i < TASK_QUEUE__END; i++) {
-			if (pp->nidle < (reserve * i / TASK_QUEUE__END))
+		for (i = 0; i < TASK_QUEUE_RESERVE; i++) {
+			if (pp->nidle < (reserve * i / TASK_QUEUE_RESERVE))
 				break;
 			tp = VTAILQ_FIRST(&pp->queues[i]);
 			if (tp != NULL) {
@@ -437,10 +436,11 @@ Pool_Work_Thread(struct pool *pp, struct worker *wrk)
 				else if (wrk->wpriv->vcl == NULL)
 					tmo = 0;
 				else if (DO_DEBUG(DBG_VTC_MODE))
-					tmo =  now + 1.;
+					tmo = now + 1.;
 				else
-					tmo =  now + 60.;
-				(void)Lck_CondWait(&wrk->cond, &pp->mtx, tmo);
+					tmo = now + 60.;
+				(void)Lck_CondWaitUntil(
+				    &wrk->cond, &pp->mtx, tmo);
 				if (wrk->task->func != NULL) {
 					/* We have been handed a new task */
 					tpx = *wrk->task;
@@ -483,6 +483,7 @@ Pool_Work_Thread(struct pool *pp, struct worker *wrk)
 		do {
 			memset(wrk->task, 0, sizeof wrk->task);
 			assert(wrk->pool == pp);
+			AN(tp->func);
 			tp->func(wrk, tp->priv);
 			if (DO_DEBUG(DBG_VCLREL) && wrk->wpriv->vcl != NULL)
 				VCL_Rel(&wrk->wpriv->vcl);
@@ -680,8 +681,9 @@ pool_herder(void *priv)
 				VSC_C_main->threads_destroyed++;
 				Lck_Unlock(&pool_mtx);
 				delay = cache_param->wthread_destroy_delay;
-			} else if (delay < cache_param->wthread_destroy_delay)
-				delay = cache_param->wthread_destroy_delay;
+			} else
+				delay = vmax(delay,
+				    cache_param->wthread_destroy_delay);
 		}
 
 		if (pp->die) {
@@ -696,14 +698,14 @@ pool_herder(void *priv)
 		if (pp->lqueue == 0) {
 			if (DO_DEBUG(DBG_VTC_MODE))
 				delay = 0.5;
-			r = Lck_CondWait(&pp->herder_cond, &pp->mtx,
-			    VTIM_real() + delay);
+			r = Lck_CondWaitTimeout(
+			    &pp->herder_cond, &pp->mtx, delay);
 		} else if (pp->nthr >= cache_param->wthread_max) {
 			/* XXX: unsafe counters */
 			if (r != ETIMEDOUT)
 				VSC_C_main->threads_limited++;
-			r = Lck_CondWait(&pp->herder_cond, &pp->mtx,
-			    VTIM_real() + 1.0);
+			r = Lck_CondWaitTimeout(
+			    &pp->herder_cond, &pp->mtx, 1.0);
 		}
 		Lck_Unlock(&pp->mtx);
 	}
@@ -735,6 +737,25 @@ static struct cli_proto debug_cmds[] = {
 	{ NULL }
 };
 
+void
+WRK_Log(enum VSL_tag_e tag, const char *fmt, ...)
+{
+	struct worker *wrk;
+	va_list ap;
+
+	AN(fmt);
+
+	wrk = THR_GetWorker();
+	CHECK_OBJ_ORNULL(wrk, WORKER_MAGIC);
+
+	va_start(ap, fmt);
+	if (wrk != NULL && wrk->vsl != NULL)
+		VSLbv(wrk->vsl, tag, fmt, ap);
+	else
+		VSLv(tag, 0, fmt, ap);
+	va_end(ap);
+}
+
 /*--------------------------------------------------------------------
  *
  */
@@ -742,6 +763,6 @@ static struct cli_proto debug_cmds[] = {
 void
 WRK_Init(void)
 {
-	assert(cache_param->wthread_min >= TASK_QUEUE__END);
+	assert(cache_param->wthread_min >= TASK_QUEUE_RESERVE);
 	CLI_AddFuncs(debug_cmds);
 }
