@@ -204,8 +204,7 @@ MCH_TrackHighFd(int fd)
 	 * in the master process.
 	 */
 	assert(fd > 0);
-	if (fd > mgt_max_fd)
-		mgt_max_fd = fd;
+	mgt_max_fd = vmax(mgt_max_fd, fd);
 }
 
 /*--------------------------------------------------------------------
@@ -293,7 +292,7 @@ child_poker(const struct vev *e, int what)
 static void
 mgt_launch_child(struct cli *cli)
 {
-	pid_t pid;
+	pid_t pid, pidr;
 	unsigned u;
 	char *p;
 	struct vev *e;
@@ -379,6 +378,10 @@ mgt_launch_child(struct cli *cli)
 		heritage.cls = mgt_cls;
 		heritage.ident = VSB_data(vident) + 1;
 
+		vext_load();
+
+		STV_Init();
+
 		VJ_subproc(JAIL_SUBPROC_WORKER);
 
 		heritage.proc_vsmw = VSMW_New(heritage.vsm_fd, 0640, "_.index");
@@ -386,7 +389,7 @@ mgt_launch_child(struct cli *cli)
 
 		/*
 		 * We pass these two params because child_main needs them
-		 * Well before it has found its own param struct.
+		 * well before it has found its own param struct.
 		 */
 		child_main(mgt_param.sigsegv_handler,
 		    mgt_param.wthread_stacksize);
@@ -418,6 +421,19 @@ mgt_launch_child(struct cli *cli)
 	child_std_vlu = VLU_New(child_line, NULL, 0);
 	AN(child_std_vlu);
 
+	/* Wait for cache/cache_cli.c::CLI_Run() to check in */
+	if (VCLI_ReadResult(child_cli_in, &u, NULL, mgt_param.cli_timeout)) {
+		assert(u == CLIS_COMMS);
+		pidr = waitpid(pid, &i, 0);
+		assert(pidr == pid);
+		(void)VLU_Fd(child_std_vlu, child_output);
+		MGT_Complain(C_ERR, "Child failed on launch");
+		exit(1);		// XXX Harsh ?
+	} else {
+		assert(u == CLIS_OK);
+		fprintf(stderr, "Child launched OK\n");
+	}
+
 	AZ(ev_listen);
 	e = VEV_Alloc();
 	XXXAN(e);
