@@ -148,6 +148,19 @@ vcc_delete_expr(struct expr *e)
  * XXX: check line lengths in edit, should pass indent in for this
  */
 
+static void
+vcc_strands_edit(const struct expr *e1, const struct expr *e2)
+{
+
+	if (e2->nstr == 1) {
+		VSB_printf(e1->vsb, "TOSTRAND(%s)", VSB_data(e2->vsb));
+		return;
+	}
+
+	VSB_printf(e1->vsb, "TOSTRANDS(%d,\v+\n%s\v-)",
+	    e2->nstr, VSB_data(e2->vsb));
+}
+
 static struct expr *
 vcc_expr_edit(struct vcc *tl, vcc_type_t fmt, const char *p, struct expr *e1,
     struct expr *e2)
@@ -179,10 +192,7 @@ vcc_expr_edit(struct vcc *tl, vcc_type_t fmt, const char *p, struct expr *e1,
 			if (e3->nstr > 1) {
 				VSB_cat(e->vsb,
 				    "\nVRT_STRANDS_string(ctx,\v+\n");
-				VSB_printf(e->vsb,
-				    "&(struct strands){.n = %d, .p = "
-				    "(const char *[%d]){\n%s\n}}",
-				    e3->nstr, e3->nstr, VSB_data(e3->vsb));
+				vcc_strands_edit(e, e3);
 				VSB_cat(e->vsb,
 				    "\v-\n)\n");
 			} else {
@@ -193,10 +203,7 @@ vcc_expr_edit(struct vcc *tl, vcc_type_t fmt, const char *p, struct expr *e1,
 		case 't':
 			e3 = (*p == 'T' ? e1 : e2);
 			AN(e3);
-			VSB_printf(e->vsb,
-			    "&(struct strands){.n = %d, .p = "
-			    "(const char *[%d]){\n%s\n}}",
-			    e3->nstr, e3->nstr, VSB_data(e3->vsb));
+			vcc_strands_edit(e, e3);
 			break;
 		case '1':
 			VSB_cat(e->vsb, VSB_data(e1->vsb));
@@ -230,8 +237,10 @@ vcc_expr_fmt(struct vsb *d, int ind, const struct expr *e1)
 	char *p;
 	int i;
 
-	for (i = 0; i < ind; i++)
-		VSB_cat(d, " ");
+	if (!e1->fmt->noindent) {
+		for (i = 0; i < ind; i++)
+			VSB_putc(d, ' ');
+	}
 	p = VSB_data(e1->vsb);
 	while (*p != '\0') {
 		if (*p == '\n') {
@@ -239,13 +248,13 @@ vcc_expr_fmt(struct vsb *d, int ind, const struct expr *e1)
 			if (*++p == '\0')
 				break;
 			for (i = 0; i < ind; i++)
-				VSB_cat(d, " ");
+				VSB_putc(d, ' ');
 		} else if (*p != '\v') {
 			VSB_putc(d, *p++);
 		} else {
 			switch (*++p) {
-			case '+': ind += 2; break;
-			case '-': ind -= 2; break;
+			case '+': ind += INDENT; break;
+			case '-': ind -= INDENT; break;
 			default:  WRONG("Illegal format in VCC expression");
 			}
 			p++;
@@ -496,7 +505,7 @@ vcc_func(struct vcc *tl, struct expr **e, const void *priv,
 	int n;
 
 	CAST_OBJ_NOTNULL(vv, priv, VJSN_VAL_MAGIC);
-	assert(vv->type == VJSN_ARRAY);
+	assert(vjsn_is_array(vv));
 	vv = VTAILQ_FIRST(&vv->children);
 	rfmt = VCC_Type(VTAILQ_FIRST(&vv->children)->value);
 	AN(rfmt);
@@ -530,7 +539,7 @@ vcc_func(struct vcc *tl, struct expr **e, const void *priv,
 	}
 	VTAILQ_INIT(&head);
 	for (;vv != NULL; vv = VTAILQ_NEXT(vv, list)) {
-		assert(vv->type == VJSN_ARRAY);
+		assert(vjsn_is_array(vv));
 		fa = calloc(1, sizeof *fa);
 		AN(fa);
 		fa->cname = cfunc;
@@ -559,7 +568,7 @@ vcc_func(struct vcc *tl, struct expr **e, const void *priv,
 				}
 			}
 		}
-		if (sa != NULL && vvp != NULL && vvp->type == VJSN_TRUE) {
+		if (sa != NULL && vvp != NULL && vjsn_is_true(vvp)) {
 			fa->optional = 1;
 			vvp = VTAILQ_NEXT(vvp, list);
 		}
@@ -780,6 +789,7 @@ vcc_expr5(struct vcc *tl, struct expr **e, vcc_type_t fmt)
 			AN(sym);
 			VCC_GlobalSymbol(sym, fmt);
 		}
+		ERRCHK(tl);
 		if (sym == NULL)
 			AZ(VCC_SymbolGet(tl, SYM_MAIN, SYM_NONE,
 			    SYMTAB_PARTIAL, XREF_REF));
@@ -969,8 +979,7 @@ vcc_expr_mul(struct vcc *tl, struct expr **e, vcc_type_t fmt)
 
 	while (tl->t->tok == '*' || tl->t->tok == '/' || tl->t->tok == '%') {
 		if (tl->t->tok == '%' && ((*e)->fmt != INT)) {
-			VSB_printf(tl->sb,
-			    "Operator %% only possible on INT.\n");
+			VSB_cat(tl->sb, "Operator % only possible on INT.\n");
 			vcc_ErrWhere(tl, tl->t);
 			return;
 		}
@@ -1423,6 +1432,18 @@ vcc_expr0(struct vcc *tl, struct expr **e, vcc_type_t fmt)
 		return;
 	}
 
+	if (fmt == BODY && !(*e)->fmt->bodyform)
+		vcc_expr_tostring(tl, e, STRINGS);
+
+	if (fmt == BODY && (*e)->fmt->bodyform) {
+		if ((*e)->fmt == STRINGS)
+			*e = vcc_expr_edit(tl, BODY, "STRING, 0, \vT", *e, NULL);
+		else if ((*e)->fmt == BLOB)
+			*e = vcc_expr_edit(tl, BODY, "BLOB, 0, \v1", *e, NULL);
+		else
+			WRONG("Unhandled bodyform");
+	}
+
 	if ((*e)->fmt == STRINGS && fmt->stringform) {
 		if (fmt == STRING)
 			*e = vcc_expr_edit(tl, STRING, "\vS", *e, NULL);
