@@ -38,6 +38,8 @@
 
 #include "cache/cache_varnishd.h"
 #include "cache/cache_filter.h"
+#include "vcc_interface.h"	// struct wrk_vpi
+
 
 #include "vsa.h"
 #include "vtim.h"
@@ -70,7 +72,7 @@ extern void mylog(struct vsl_log *vsl, enum VSL_tag_e tag,
 /**********************************************************************/
 
 static enum vfp_status v_matchproto_(vfp_pull_f)
-xyzzy_rot13_pull(struct vfp_ctx *vc, struct vfp_entry *vfe, void *p,
+xyzzy_vfp_rot13_pull(struct vfp_ctx *vc, struct vfp_entry *vfe, void *p,
     ssize_t *lp)
 {
 	enum vfp_status vp;
@@ -91,9 +93,9 @@ xyzzy_rot13_pull(struct vfp_ctx *vc, struct vfp_entry *vfe, void *p,
 	return (vp);
 }
 
-static const struct vfp xyzzy_rot13 = {
+static const struct vfp xyzzy_vfp_rot13 = {
 	.name = "rot13",
-	.pull = xyzzy_rot13_pull,
+	.pull = xyzzy_vfp_rot13_pull,
 };
 
 /**********************************************************************/
@@ -102,10 +104,12 @@ static const struct vfp xyzzy_rot13 = {
 #define ROT13_BUFSZ 8
 
 static int v_matchproto_(vdp_init_f)
-xyzzy_rot13_init(struct vdp_ctx *vdc, void **priv, struct objcore *oc)
+xyzzy_vfp_rot13_init(VRT_CTX, struct vdp_ctx *vdc, void **priv, struct objcore *oc)
 {
 	(void)vdc;
 	(void)oc;
+
+	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	AN(priv);
 	*priv = malloc(ROT13_BUFSZ);
 	if (*priv == NULL)
@@ -114,7 +118,7 @@ xyzzy_rot13_init(struct vdp_ctx *vdc, void **priv, struct objcore *oc)
 }
 
 static int v_matchproto_(vdp_bytes_f)
-xyzzy_rot13_bytes(struct vdp_ctx *vdx, enum vdp_action act, void **priv,
+xyzzy_vfp_rot13_bytes(struct vdp_ctx *vdx, enum vdp_action act, void **priv,
     const void *ptr, ssize_t len)
 {
 	char *q;
@@ -152,7 +156,7 @@ xyzzy_rot13_bytes(struct vdp_ctx *vdx, enum vdp_action act, void **priv,
 }
 
 static int v_matchproto_(vdp_fini_f)
-xyzzy_rot13_fini(struct vdp_ctx *vdc, void **priv)
+xyzzy_vfp_rot13_fini(struct vdp_ctx *vdc, void **priv)
 {
 	(void)vdc;
 	AN(priv);
@@ -163,9 +167,9 @@ xyzzy_rot13_fini(struct vdp_ctx *vdc, void **priv)
 
 static const struct vdp xyzzy_vdp_rot13 = {
 	.name  = "rot13",
-	.init  = xyzzy_rot13_init,
-	.bytes = xyzzy_rot13_bytes,
-	.fini  = xyzzy_rot13_fini,
+	.init  = xyzzy_vfp_rot13_init,
+	.bytes = xyzzy_vfp_rot13_bytes,
+	.fini  = xyzzy_vfp_rot13_fini,
 };
 
 /**********************************************************************
@@ -451,10 +455,17 @@ event_load(VRT_CTX, struct vmod_priv *priv)
 	priv->priv = priv_vcl;
 	priv->methods = priv_vcl_methods;
 
-	VRT_AddVFP(ctx, &xyzzy_rot13);
+	AZ(VRT_AddFilter(ctx, &xyzzy_vfp_rot13, &xyzzy_vdp_rot13));
 
-	VRT_AddVDP(ctx, &xyzzy_vdp_rot13);
-	VRT_AddVDP(ctx, &xyzzy_vdp_pedantic);
+	// This should fail
+	AN(VRT_AddFilter(ctx, &xyzzy_vfp_rot13, &xyzzy_vdp_rot13));
+	// Reset the error, we know what we're doing.
+	ctx->vpi->handling = 0;
+
+	VRT_RemoveFilter(ctx, &xyzzy_vfp_rot13, &xyzzy_vdp_rot13);
+	AZ(VRT_AddFilter(ctx, &xyzzy_vfp_rot13, &xyzzy_vdp_rot13));
+
+	AZ(VRT_AddFilter(ctx, NULL, &xyzzy_vdp_pedantic));
 	return (0);
 }
 
@@ -517,15 +528,10 @@ event_warm(VRT_CTX, const struct vmod_priv *priv)
 {
 	struct priv_vcl *priv_vcl;
 	char buf[32];
-	const char *p[2];
-	struct strands msg[1];
+	const char *vcl_name = VCL_Name(ctx->vcl);
 
 	// Using VSLs for coverage
-	msg->n = 2;
-	msg->p = p;
-	p[0] = VCL_Name(ctx->vcl);
-	p[1] = ": VCL_EVENT_WARM";
-	VSLs(SLT_Debug, 0, msg);
+	VSLs(SLT_Debug, 0, TOSTRANDS(2, vcl_name, ": VCL_EVENT_WARM"));
 
 	AN(ctx->msg);
 	if (cache_param->max_esi_depth == 42) {
@@ -621,9 +627,8 @@ event_discard(VRT_CTX, void *priv)
 
 	AZ(ctx->msg);
 
-	VRT_RemoveVFP(ctx, &xyzzy_rot13);
-	VRT_RemoveVDP(ctx, &xyzzy_vdp_rot13);
-	VRT_RemoveVDP(ctx, &xyzzy_vdp_pedantic);
+	VRT_RemoveFilter(ctx, &xyzzy_vfp_rot13, &xyzzy_vdp_rot13);
+	VRT_RemoveFilter(ctx, NULL, &xyzzy_vdp_pedantic);
 
 	if (--loads)
 		return (0);
@@ -810,7 +815,8 @@ xyzzy_sethdr(VRT_CTX, VCL_HEADER hdr, VCL_STRANDS s)
 	} else {
 		b = VRT_StrandsWS(hp->ws, hdr->what + 1, s);
 		if (b == NULL) {
-			VSLb(ctx->vsl, SLT_LostHeader, "%s", hdr->what + 1);
+			VSLbs(ctx->vsl, SLT_LostHeader,
+			    TOSTRAND(hdr->what + 1));
 		} else {
 			if (*b != '\0')
 				AN(WS_Allocated(hp->ws, b, strlen(b) + 1));
@@ -1052,55 +1058,6 @@ xyzzy_get_ip(VRT_CTX)
 	return (ip);
 }
 
-/**********************************************************************
- * For testing import code on bad vmod files (m00003.vtc)
- */
-
-//lint -save -e9075 external symbol '...' defined without a prior declaration
-
-extern const struct vmod_data Vmod_wrong0_Data;
-const struct vmod_data Vmod_wrong0_Data = {
-	.vrt_major =    0,
-	.vrt_minor =    0,
-};
-
-//lint -save -e835 A zero has been given as left argument to operatorp'+'
-extern const struct vmod_data Vmod_wrong1_Data;
-const struct vmod_data Vmod_wrong1_Data = {
-	.vrt_major =    VRT_MAJOR_VERSION,
-	.vrt_minor =    VRT_MINOR_VERSION + 1,
-};
-//lint -restore
-
-static const struct foo {
-	int bar;
-} foo_struct[1];
-
-extern const struct vmod_data Vmod_wrong2_Data;
-const struct vmod_data Vmod_wrong2_Data = {
-	.vrt_major =    VRT_MAJOR_VERSION,
-	.vrt_minor =    VRT_MINOR_VERSION,
-	.name =		"wrongN",
-	.func =		foo_struct,
-	.func_len =	sizeof foo_struct,
-	.func_name =	"foo_struct",
-	.proto =	"blablabla",
-};
-
-extern const struct vmod_data Vmod_wrong3_Data;
-const struct vmod_data Vmod_wrong3_Data = {
-	.vrt_major =    VRT_MAJOR_VERSION,
-	.vrt_minor =    VRT_MINOR_VERSION,
-	.name =		"wrongN",
-	.func =		foo_struct,
-	.func_len =	sizeof foo_struct,
-	.func_name =	"foo_struct",
-	.proto =	"blablabla",
-	.abi =		"abiblabla",
-};
-
-//lint -restore
-
 /*---------------------------------------------------------------------*/
 
 struct VPFX(debug_director) {
